package v800;

import java.io.*;
import java.util.*;

/**
 * Title:        GRE Verbal Reviewer
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:      Freeware
 * @author Charlie Hsuan-tien Lin
 * @version 1.0
 */

public class Main {
  private static final String wordFile = "WordList";
  private static DataContainer dc;

  private static void readData(String fileName) throws FileNotFoundException, IOException{
    dc = new DataContainer(fileName + ".txt", fileName + ".sta");
  }

  static CommandLineMenu mainMenu = null;
  static{
    Vector v = new Vector();
    v.add(new CommandLineMenuKeyString('0', "Review"));
    v.add(new CommandLineMenuKeyString('1', "Test"));
    v.add(new CommandLineMenuKeyString('9', "Quit"));
    mainMenu = new CommandLineMenu(v);
  }

  private static void selectMenu(){
    InputStream in = System.in;
    int c;
    while(true){
      switch(mainMenu.input()){
      case '0':
        dc.reviewAll();
        break;
      case '1':
        dc.testAll();
        break;
      case '9':
        return;
      }
    }
  }

  private static void writeData() throws IOException{
    dc.saveAll();
  }

  public static void main(String[] args) {
    String fileName = wordFile;
    if (args.length > 0){
      fileName = args[0];
    }
    try{
      readData(fileName);
      selectMenu();
      writeData();
    }
    catch(Exception e){
      e.printStackTrace();
    }
  }
}

class DataContainer{
  private static final int MAXLEVEL = 6;
  private static final int MAXSELECT = 5;
  private final Random rd = new Random();
  private String statFile = null;
  private String wordFile = null;
  private long total = 0;
  private Vector[] arr = new Vector[MAXLEVEL];
  {
    for(int i=0;i<MAXLEVEL;i++)
      arr[i] = new Vector();
  }

  private long weightOf(int i){
    return  1 << (MAXLEVEL-i);
  }

  private void addElement(String line, int c){
    c = (c <= 0 ? 0 : c);
    c = (c >= MAXLEVEL-1 ? MAXLEVEL-1 : c);
    arr[c].add(line);
  }

  public DataContainer(String _wordFile, String _statFile) throws FileNotFoundException, IOException{
    BufferedReader fw;
    FileReader fs;
    String line;
    int c;

    wordFile = _wordFile;
    statFile = _statFile;
    try{
      fw = new BufferedReader(new FileReader(wordFile));
    }
    catch(FileNotFoundException e){
      throw e;
    }

    try{
      fs = new FileReader(statFile);
    }
    catch(FileNotFoundException e){
      fs = null;
    }

    while(true){
      try{
        line = fw.readLine();
        if (line == null)
          break;
        if (fs == null)
          c = 0;
        else
          c = fs.read();
        addElement(line, c);
      }
      catch(IOException e){
        throw e;
      }
    }

    for(int i=0;i<MAXLEVEL;i++){
      total += weightOf(i) * arr[i].size();
    }
  }

  public void saveAll() throws IOException{
    PrintWriter fw;
    BufferedOutputStream fs;
    try{
      fw = new PrintWriter(new BufferedWriter(new FileWriter(wordFile)));
      fs = new BufferedOutputStream(new FileOutputStream(statFile));

      for(int i=0;i<MAXLEVEL;i++){
        Object[] tmp = arr[i].toArray();
        Arrays.sort(tmp);
        for(int j=0;j<tmp.length;j++){
          fw.println((String)tmp[j]);
          fs.write(i);
        }
      }
      fw.flush();
      fs.flush();
    }
    catch(IOException e){
      throw e;
    }
  }

  public String getItem(long index){
    try{
      return (String)(arr[(int)(index >> 32)].elementAt((int)index));
    }
    catch(Exception e){
      return null;
    }
  }

  public void liftUp(long index){
    int level = (int)(index >> 32);
    int num = (int)index;
    if (level < MAXLEVEL){
      Object o = arr[level].elementAt(num);
      arr[level].removeElementAt(num);
      total -= weightOf(level);
      level++;
      arr[level].add(o);
      total += weightOf(level);
    }
  }

  public void pushDown(long index){
    int level = (int)(index >> 32);
    int num = (int)index;
    if (level > 0){
      Object o = arr[level].elementAt(num);
      arr[level].removeElementAt(num);
      total -= weightOf(level);
      level--;
      arr[level].add(o);
      total += weightOf(level);
    }
  }

  public long chooseByProb() throws Exception{
    long tofind = rd.nextLong();
    tofind = (tofind > 0 ? tofind : -tofind);
    tofind %= total;
    for(int i=0;i<MAXLEVEL;i++){
      long weight = weightOf(i);
      if (tofind/weight < arr[i].size()){
        return ((long)i << 32) + (int)(tofind / weight);
      }
      else
        tofind -= weight * arr[i].size();
    }
    throw new Exception("Statement should not be reached");
  }

  private class PairString{
    public String first, second;
    public long index;
    public PairString(DataContainer dc, long _index) throws Exception{
      String line = dc.getItem(_index);
      StringTokenizer st = new StringTokenizer(line);
      index = _index;
      try{
        first = st.nextToken(",");
        second = st.nextToken(",\n\r\f");
      }
      catch(Exception e){
        throw e;
      }
    }
  }

  public void reviewAll(){
    while(reviewOne());
  }

  CommandLineMenu reviewOneMenu = null;
  {
    Vector v = new Vector();
    v.add(new CommandLineMenuKeyString('0', "See the meaning"));
    v.add(new CommandLineMenuKeyString('9', "Quit"));
    reviewOneMenu = new CommandLineMenu(v);
  }

  public boolean reviewOne(){
    long id = -1;
    while(true){
      try{
        id = chooseByProb();
        break;
      }
      catch(Exception e){
        e.printStackTrace();
        continue;
      }
    }

    PairString ps;
    try{
      ps = new PairString(this, id);
    }
    catch(Exception e){
      return false;
    }

    System.out.println(ps.first);
    if (reviewOneMenu.input() == '9'){
      System.out.println(ps.second);
      return false;
    }
    else{
      System.out.println(ps.second);
      return true;
    }
  }

  public void showStatus(){
    System.out.println("=====Current Status=====");
    for(int i=MAXLEVEL-1;i>=0;i--)
      System.out.println("Class " + i + ": " + arr[i].size());
    System.out.println("========================");
  }

  public void testAll(){
    do{
      showStatus();
    }while(testOne());
  }

  CommandLineMenu testOneMenu = null;
  {
    Vector v = new Vector();
    testOneMenu = new CommandLineMenu(v);
  }

  private boolean isConflict(PairString[] psarr, int index){
    for(int i=index-1;i>=0;i--){
      if (psarr[i].index == psarr[index].index)
        return true;
      else if (psarr[i].first.equals(psarr[index].first))
        return true;
    }
    return false;
  }

  private int permute(PairString[] psarr){
    int ans = 0;
    for(int i=0;i<MAXSELECT;i++){
      PairString tmp;
      int index = rd.nextInt(MAXSELECT-i) + i;
      tmp = psarr[i];
      psarr[i] = psarr[index];
      psarr[index] = tmp;
      if (i == ans)
        ans = index;
      else if (index == ans)
        ans = i;
    }
    return ans;
  }

  public boolean testOne(){
    PairString[] problem = new PairString[MAXSELECT];
    int ans = 0;

    for(int i=0;i<MAXSELECT;i++){
      while(true){
        try{
          problem[i] = new PairString(this, chooseByProb());
          if (!isConflict(problem, i))
            break;
        }
        catch(Exception e){
          e.printStackTrace();
          continue;
        }
      }
    }

    ans = permute(problem);
    testOneMenu.clearAll();
    testOneMenu.insertFront(new CommandLineMenuKeyString('9', "Quit"));
    testOneMenu.insertFront(new CommandLineMenuKeyString('8', "I don't know"));

    for(int i=MAXSELECT-1;i>=0;i--){
      testOneMenu.insertFront(new CommandLineMenuKeyString((char)('0' + i), problem[i].second));
    }

    System.out.println(problem[ans].first);
    char c = testOneMenu.input();
    if (c == '9'){
      return false;
    }
    else if (c == ans + '0'){
      System.out.println("Correct!!");
      liftUp(problem[ans].index);
      return true;
    }
    else{
      System.out.println("Wrong, the answer is:");
      System.out.print(problem[ans].first);
      System.out.print("==>");
      System.out.println(problem[ans].second);
      testOneMenu.pressEnter();
      pushDown(problem[ans].index);
      if (c < MAXSELECT + '0')
        pushDown(problem[c - '0'].index);
      return true;
    }
  }
}

class CommandLineMenuKeyString{
  public char key;
  public String toShow;

  public CommandLineMenuKeyString(char _key, String _toShow){
    key = _key;
    toShow = _toShow;
  }
}

class CommandLineMenu{
  private Vector keylist;
  private BufferedReader in = new  BufferedReader(new InputStreamReader(System.in));

  public CommandLineMenu(Vector _keylist){
    keylist = _keylist;
    for(int i=0;i<keylist.size();i++){
      if (keylist.elementAt(i) instanceof CommandLineMenuKeyString){
        continue;
      }
      else{
        keylist.removeElementAt(i);
        i--;
      }
    }
  }

  public void insertFront(CommandLineMenuKeyString item){
    keylist.add(0, item);
  }

  public void clearAll(){
    keylist.clear();
  }

  public void pressEnter(){
    BufferedReader in = this.in;
    System.out.println("Press Enter to Continue");
    try{
      in.readLine();
    }
    catch(Exception e){
      e.printStackTrace();
    }
 }

  public char input(){
    PrintStream out = System.out;
    BufferedReader in = this.in;

    for(int i=0;i<keylist.size();i++){
      CommandLineMenuKeyString item = (CommandLineMenuKeyString)keylist.elementAt(i);
      out.print("(");
      out.print(item.key);
      out.print(") ");
      out.println(item.toShow);
    }
    while(true){
      String s;
      int c;
      try{
        s = in.readLine();
        c = s.charAt(0);
      }
      catch(Exception e){
        c = -1;
      }

      switch(c){
      case -1:
        continue;
      default:
        for(int i=0;i<keylist.size();i++){
          CommandLineMenuKeyString item = (CommandLineMenuKeyString)keylist.elementAt(i);
          if (item.key == c)
            return item.key;
        }
      }
    }
  }
}
